unit FreeOTFESettings;
// Description: 
// By Sarah Dean
// Email: sdean12@sdean12.org
// WWW:   http://www.FreeOTFE.org/
//
// -----------------------------------------------------------------------------
//


interface

uses
  Classes, // Required for TShortCut
  IniFiles,
  CommonSettings;

{$IFDEF _NEVER_DEFINED}
// This is just a dummy const to fool dxGetText when extracting message
// information
// This const is never used; it's #ifdef'd out - SDUCRLF in the code refers to
// picks up SDUGeneral.SDUCRLF
const
  SDUCRLF = ''#13#10;
{$ENDIF}
  
const
  FREEOTFE_REGISTRY_SETTINGS_LOCATION = '\Software\FreeOTFE';
  
  DragDropFileTypeTitlePtr: array [TDragDropFileType] of Pointer = (
                                  @DRAGDROPFILETYPE_PROMPT,
                                  @DRAGDROPFILETYPE_FREEOTFE,
                                  @DRAGDROPFILETYPE_LINUX
                                 );

type
  TFreeOTFESettings = class(TSettings)
  protected
    procedure _Load(iniFile: TCustomINIFile); override;
    function  _Save(iniFile: TCustomINIFile): boolean; override;

  public
    // General...
    OptDisplayToolbar: boolean;
    OptDisplayToolbarLarge: boolean;
    OptDisplayToolbarCaptions: boolean;
    OptExploreAfterMount: boolean;
    OptAllowMultipleInstances: boolean;
    OptAutoStartPortable: boolean;
    OptDragDropFileType: TDragDropFileType;
    OptDefaultDriveLetter: char;

    // Prompts and messages
    OptPromptMountSuccessful: boolean;  // If set, display an info msg after successful mount
    OptPromptDismountOnClose: boolean;  // If set, prompt the user if they want to dismount any mounted volumes when closing
    OptPromptDriverShutdownOnClose: boolean;  // If set, prompt the user if they want to shutdown any portable mode drivers when closing

    // System tray icon...
    OptSystemTrayIconDisplay: boolean;
    OptSystemTrayIconMinTo: boolean;
    OptSystemTrayIconCloseTo: boolean;

    // Hotkey...
    OptHKeyEnableDismount: boolean;
    OptHKeyKeyDismount: TShortCut;
    OptHKeyEnableDismountEmerg: boolean;
    OptHKeyKeyDismountEmerg: TShortCut;

    // Autorun...
    OptPostMountExe: string;
    OptPreDismountExe: string;
    OptPostDismountExe: string;
    OptPrePostExeWarn: boolean;

    constructor Create(); override;
    destructor Destroy(); override;

    function  RegistryKey(): string; override;

  end;

  function DragDropFileTypeTitle(dragDropfileType: TDragDropFileType): string;

var
  // Global variable
  Settings: TFreeOTFESettings;

implementation

uses
  Windows,  // Required to get rid of compiler hint re DeleteFile
  SysUtils,  // Required for ChangeFileExt, DeleteFile
  Registry,
  Dialogs,
  SDUi18n,
  SDUGeneral,
  SDUDialogs,
  Menus,  // Required for ShortCutToText and TextToShortCut
  ShlObj;  // Required for CSIDL_PERSONAL

const
  SETTINGS_V1 = 1;
  SETTINGS_V2 = 2;

  // -- General section --
    OPT_DISPLAYTOOLBAR                           = 'DisplayToolbar';
      DFLT_OPT_DISPLAYTOOLBAR                       = TRUE;
    OPT_DISPLAYTOOLBARLARGE                      = 'DisplayToolbarLarge';
      DFLT_OPT_DISPLAYTOOLBARLARGE                  = TRUE;
    OPT_DISPLAYTOOLBARCAPTIONS                   = 'DisplayToolbarCaptions';
      DFLT_OPT_DISPLAYTOOLBARCAPTIONS               = TRUE;
    OPT_EXPLOREAFTERMOUNT                        = 'ExploreAfterMount';
      DFLT_OPT_EXPLOREAFTERMOUNT                    = FALSE;
    OPT_ALLOWMULTIPLEINSTANCES                   = 'AllowMultipleInstances';
      DFLT_OPT_ALLOWMULTIPLEINSTANCES               = FALSE;
    OPT_AUTOSTARTPORTABLE                        = 'AutoStartPortable';
      DFLT_OPT_AUTOSTARTPORTABLE                    = FALSE;
    OPT_DRAGDROP                                 = 'DragDropFileType';
      DFLT_OPT_DRAGDROP                             = ord(ftFreeOTFE);
    OPT_DEFAULTDRIVELETTER                       = 'DefaultDriveLetter';
      DFLT_OPT_DEFAULTDRIVELETTER                   = '#';

  // -- Prompts and messages --
  SECTION_CONFIRMATION = 'Confirmation';
    OPT_PROMPTMOUNTSUCCESSFUL                    = 'OptPromptMountSuccessful';
      DFLT_OPT_PROMPTMOUNTSUCCESSFUL                = TRUE;
    OPT_PROMPTDISMOUNTONCLOSE                    = 'OptPromptDismountOnClose';
      DFLT_OPT_PROMPTDISMOUNTONCLOSE                = TRUE;
    OPT_PROMPTDRIVERSHUTDOWNONCLOSE              = 'OptPromptDriverShutdownOnClose';
      DFLT_OPT_PROMPTDRIVERSHUTDOWNONCLOSE          = TRUE;

  // -- System tray icon section --
  SECTION_SYSTEMTRAYICON = 'SystemTrayIcon';
    OPT_SYSTEMTRAYICONDISPLAY                    = 'Display';
      DFLT_OPT_SYSTEMTRAYICONDISPLAY                = TRUE;
    OPT_SYSTEMTRAYICONMINTO                      = 'MinTo';
      DFLT_OPT_SYSTEMTRAYICONMINTO                  = FALSE;
    OPT_SYSTEMTRAYICONCLOSETO                    = 'CloseTo';
      DFLT_OPT_SYSTEMTRAYICONCLOSETO                = FALSE;

  // -- Hotkey section --
  SECTION_HOTKEY = 'Hotkeys';
    OPT_HOTKEYENABLEDISMOUNT                     = 'EnableDismount';
      DFLT_OPT_HOTKEYENABLEDISMOUNT                 = FALSE;
    OPT_HOTKEYKEYDISMOUNT                        = 'KeyDismount';
      DFLT_OPT_HOTKEYKEYDISMOUNT                    = 'Shift+Ctrl+D';
    OPT_HOTKEYENABLEDISMOUNTEMERG                = 'EnableDismountEmerg';
      DFLT_OPT_HOTKEYENABLEDISMOUNTEMERG            = FALSE;
    OPT_HOTKEYKEYDISMOUNTEMERG                   = 'KeyDismountEmerg';
      DFLT_OPT_HOTKEYKEYDISMOUNTEMERG               = 'Shift+Ctrl+Alt+D';

  // -- Autorun section --
  SECTION_AUTORUN = 'Autorun';
    OPT_POSTMOUNTEXE                             = 'PostMountExe';
      DFLT_OPT_POSTMOUNTEXE                         = '';
    OPT_PREDISMOUNTEXE                           = 'PreDismountExe';
      DFLT_OPT_PREDISMOUNTEXE                       = '';
    OPT_POSTDISMOUNTEXE                          = 'PostDismountExe';
      DFLT_OPT_POSTDISMOUNTEXE                      = '';
    OPT_PREPOSTEXEWARN                           = 'PrePostExeWarn';
      DFLT_OPT_PREPOSTEXEWARN                       = TRUE;


function DragDropFileTypeTitle(dragDropfileType: TDragDropFileType): string;
begin
  Result := LoadResString(DragDropFileTypeTitlePtr[dragDropfileType]);
end;

constructor TFreeOTFESettings.Create();
begin
  inherited;

  Load();

end;

destructor TFreeOTFESettings.Destroy();
begin

  inherited;
end;

procedure TFreeOTFESettings._Load(iniFile: TCustomINIFile);
var
  useDefaultDriveLetter: string;
begin
  inherited _Load(iniFile);

  OptDisplayToolbar          := iniFile.ReadBool(SECTION_GENERAL,   OPT_DISPLAYTOOLBAR,         DFLT_OPT_DISPLAYTOOLBAR);
  OptDisplayToolbarLarge     := iniFile.ReadBool(SECTION_GENERAL,   OPT_DISPLAYTOOLBARLARGE,    DFLT_OPT_DISPLAYTOOLBARLARGE);
  OptDisplayToolbarCaptions  := iniFile.ReadBool(SECTION_GENERAL,   OPT_DISPLAYTOOLBARCAPTIONS, DFLT_OPT_DISPLAYTOOLBARCAPTIONS);
  OptExploreAfterMount       := iniFile.ReadBool(SECTION_GENERAL,   OPT_EXPLOREAFTERMOUNT,      DFLT_OPT_EXPLOREAFTERMOUNT);
  OptAllowMultipleInstances  := iniFile.ReadBool(SECTION_GENERAL,   OPT_ALLOWMULTIPLEINSTANCES, DFLT_OPT_ALLOWMULTIPLEINSTANCES);
  OptAutoStartPortable       := iniFile.ReadBool(SECTION_GENERAL,   OPT_AUTOSTARTPORTABLE,      DFLT_OPT_AUTOSTARTPORTABLE);
  OptDragDropFileType        := TDragDropFileType(iniFile.ReadInteger(SECTION_GENERAL, OPT_DRAGDROP, DFLT_OPT_DRAGDROP));
  useDefaultDriveLetter      := iniFile.ReadString(SECTION_GENERAL, OPT_DEFAULTDRIVELETTER,     DFLT_OPT_DEFAULTDRIVELETTER);
  // #0 written as "#"
  OptDefaultDriveLetter := useDefaultDriveLetter[1];
  if (OptDefaultDriveLetter = '#') then
    begin
    OptDefaultDriveLetter := #0;
    end;

  OptPromptMountSuccessful       := iniFile.ReadBool(SECTION_CONFIRMATION, OPT_PROMPTMOUNTSUCCESSFUL,       DFLT_OPT_PROMPTMOUNTSUCCESSFUL);
  OptPromptDismountOnClose       := iniFile.ReadBool(SECTION_CONFIRMATION, OPT_PROMPTDISMOUNTONCLOSE,       DFLT_OPT_PROMPTDISMOUNTONCLOSE);
  OptPromptDriverShutdownOnClose := iniFile.ReadBool(SECTION_CONFIRMATION, OPT_PROMPTDRIVERSHUTDOWNONCLOSE, DFLT_OPT_PROMPTDRIVERSHUTDOWNONCLOSE);

  OptSystemTrayIconDisplay   := iniFile.ReadBool(SECTION_SYSTEMTRAYICON, OPT_SYSTEMTRAYICONDISPLAY, DFLT_OPT_SYSTEMTRAYICONDISPLAY);
  OptSystemTrayIconMinTo     := iniFile.ReadBool(SECTION_SYSTEMTRAYICON, OPT_SYSTEMTRAYICONMINTO,   DFLT_OPT_SYSTEMTRAYICONMINTO);
  OptSystemTrayIconCloseTo   := iniFile.ReadBool(SECTION_SYSTEMTRAYICON, OPT_SYSTEMTRAYICONCLOSETO, DFLT_OPT_SYSTEMTRAYICONCLOSETO);

  OptHKeyEnableDismount      := iniFile.ReadBool(SECTION_HOTKEY,   OPT_HOTKEYENABLEDISMOUNT,      DFLT_OPT_HOTKEYENABLEDISMOUNT);
  OptHKeyKeyDismount         := TextToShortCut(iniFile.ReadString(SECTION_HOTKEY, OPT_HOTKEYKEYDISMOUNT,         DFLT_OPT_HOTKEYKEYDISMOUNT));
  OptHKeyEnableDismountEmerg := iniFile.ReadBool(SECTION_HOTKEY,   OPT_HOTKEYENABLEDISMOUNTEMERG, DFLT_OPT_HOTKEYENABLEDISMOUNTEMERG);
  OptHKeyKeyDismountEmerg    := TextToShortCut(iniFile.ReadString(SECTION_HOTKEY, OPT_HOTKEYKEYDISMOUNTEMERG,    DFLT_OPT_HOTKEYKEYDISMOUNTEMERG));

  OptPostMountExe            := iniFile.ReadString(SECTION_AUTORUN,      OPT_POSTMOUNTEXE,          DFLT_OPT_POSTMOUNTEXE);
  OptPreDismountExe          := iniFile.ReadString(SECTION_AUTORUN,      OPT_PREDISMOUNTEXE,        DFLT_OPT_PREDISMOUNTEXE);
  OptPostDismountExe         := iniFile.ReadString(SECTION_AUTORUN,      OPT_POSTDISMOUNTEXE,       DFLT_OPT_POSTDISMOUNTEXE);
  OptPrePostExeWarn          := iniFile.ReadBool(SECTION_AUTORUN,        OPT_PREPOSTEXEWARN,        DFLT_OPT_PREPOSTEXEWARN);

end;


function TFreeOTFESettings._Save(iniFile: TCustomINIFile): boolean;
var
  allOK: boolean;
  useDefaultDriveLetter: char;
begin
  allOK := inherited _Save(iniFile);
  try
    iniFile.WriteBool(SECTION_GENERAL,        OPT_DISPLAYTOOLBAR,            OptDisplayToolbar);
    iniFile.WriteBool(SECTION_GENERAL,        OPT_DISPLAYTOOLBARLARGE,       OptDisplayToolbarLarge);
    iniFile.WriteBool(SECTION_GENERAL,        OPT_DISPLAYTOOLBARCAPTIONS,    OptDisplayToolbarCaptions);
    iniFile.WriteBool(SECTION_GENERAL,        OPT_EXPLOREAFTERMOUNT,         OptExploreAfterMount);
    iniFile.WriteBool(SECTION_GENERAL,        OPT_ALLOWMULTIPLEINSTANCES,    OptAllowMultipleInstances);
    iniFile.WriteBool(SECTION_GENERAL,        OPT_AUTOSTARTPORTABLE,         OptAutoStartPortable);
    iniFile.WriteInteger(SECTION_GENERAL,     OPT_DRAGDROP,                  ord(OptDragDropFileType));
    // #0 written as "#"
    useDefaultDriveLetter := OptDefaultDriveLetter;
    if (OptDefaultDriveLetter = #0) then
      begin
      useDefaultDriveLetter := '#';
      end;
    iniFile.WriteString(SECTION_GENERAL,      OPT_DEFAULTDRIVELETTER,        useDefaultDriveLetter);

    iniFile.WriteBool(SECTION_CONFIRMATION,  OPT_PROMPTMOUNTSUCCESSFUL,       OptPromptMountSuccessful);
    iniFile.WriteBool(SECTION_CONFIRMATION,  OPT_PROMPTDISMOUNTONCLOSE,       OptPromptDismountOnClose);
    iniFile.WriteBool(SECTION_CONFIRMATION,  OPT_PROMPTDRIVERSHUTDOWNONCLOSE, OptPromptDriverShutdownOnClose);

    iniFile.WriteBool(SECTION_SYSTEMTRAYICON, OPT_SYSTEMTRAYICONDISPLAY,     OptSystemTrayIconDisplay);
    iniFile.WriteBool(SECTION_SYSTEMTRAYICON, OPT_SYSTEMTRAYICONMINTO,       OptSystemTrayIconMinTo);
    iniFile.WriteBool(SECTION_SYSTEMTRAYICON, OPT_SYSTEMTRAYICONCLOSETO,     OptSystemTrayIconCloseTo);

    iniFile.WriteBool(SECTION_HOTKEY,         OPT_HOTKEYENABLEDISMOUNT,      OptHKeyEnableDismount);
    iniFile.WriteString(SECTION_HOTKEY,       OPT_HOTKEYKEYDISMOUNT,         ShortCutToText(OptHKeyKeyDismount));
    iniFile.WriteBool(SECTION_HOTKEY,         OPT_HOTKEYENABLEDISMOUNTEMERG, OptHKeyEnableDismountEmerg);
    iniFile.WriteString(SECTION_HOTKEY,       OPT_HOTKEYKEYDISMOUNTEMERG,    ShortCutToText(OptHKeyKeyDismountEmerg));

    iniFile.WriteString(SECTION_AUTORUN,      OPT_POSTMOUNTEXE,              OptPostMountExe);
    iniFile.WriteString(SECTION_AUTORUN,      OPT_PREDISMOUNTEXE,            OptPreDismountExe);
    iniFile.WriteString(SECTION_AUTORUN,      OPT_POSTDISMOUNTEXE,           OptPostDismountExe);
    iniFile.WriteBool(SECTION_AUTORUN,        OPT_PREPOSTEXEWARN,            OptPrePostExeWarn);

  except
    on E:Exception do
      begin
      allOK := FALSE;
      end;
  end;

  Result := allOK;
end;

function TFreeOTFESettings.RegistryKey(): string; 
begin
  Result := FREEOTFE_REGISTRY_SETTINGS_LOCATION;
end;

END.


