// Description: FreeOTFE Device Driver API
// By Sarah Dean
// Email: sdean12@sdean12.org
// WWW:   http://www.FreeOTFE.org/
//
// -----------------------------------------------------------------------------
//


#ifndef _FreeOTFE4PDAAPI_H
#define _FreeOTFE4PDAAPI_H   1


// Required for definition of ULONG
#include <windef.h>

#include "FreeOTFEAPITypes.h"


// =========================================================================
// Const definitions

// Name which may not be used as the mountpoint
// (From help IOCTL_DISK_GETNAME other language systems may have a
// different reserved name)
#define RESERVED_MOUNTPOINT           TEXT("Storage Card")
// Characters that may not appear in a mountpoint
#define RESERVED_MOUNTPOINT_CHARS     TEXT("\\/*?\"<>|")

#define FAILED_GETFILEATTRIBUTES  0xFFFFFFFF
#define FAILED_SETFILEPOINTER     0xFFFFFFFF

// Min volume file size; in MB
#define MIN_VOLUME_SIZE           2
// Max is arbitary
#define MAX_VOLUME_SIZE           9999  

// INI filenames
#define FREEOTFE_INI                 TEXT("FreeOTFE4PDA.ini")

// DLL filenames
#define DRIVER_SUBDIR                TEXT("dll\\")
#define FREEOTFE_DLL_MAIN            DRIVER_SUBDIR TEXT("FreeOTFE4PDA.dll")
#define FREEOTFE_DLL_PATTERN_CYPHER  DRIVER_SUBDIR TEXT("FreeOTFE4PDACypher*.dll")
#define FREEOTFE_DLL_PATTERN_HASH    DRIVER_SUBDIR TEXT("FreeOTFE4PDAHash*.dll")


// =========================================================================
// Exported functions...

#define DLLEXPORT_MAIN_DERIVEKEY TEXT("MainDeriveKey")
typedef DWORD (* PMainDLLFnDeriveKey)(
    IN  DWORD SizeBufferIn,
    IN  DIOC_DERIVE_KEY_IN*   DIOCBufferIn,
    IN  DWORD SizeBufferOut,
    OUT DIOC_DERIVE_KEY_OUT*  DIOCBufferOut
);
DWORD
MainDeriveKey(
    IN  DWORD SizeBufferIn,
    IN  DIOC_DERIVE_KEY_IN*   DIOCBufferIn,
    IN  DWORD SizeBufferOut,
    OUT DIOC_DERIVE_KEY_OUT*  DIOCBufferOut
);


#define DLLEXPORT_MAIN_MACDATA TEXT("MainMACData")
typedef DWORD (* PMainDLLFnMACData)(
    IN  DWORD SizeBufferIn,
    IN  DIOC_GENERATE_MAC_IN*   DIOCBufferIn,
    IN  DWORD SizeBufferOut,
    OUT DIOC_GENERATE_MAC_OUT*  DIOCBufferOut
);
DWORD
MainMACData(
    IN  DWORD SizeBufferIn,
    IN  DIOC_GENERATE_MAC_IN*   DIOCBufferIn,
    IN  DWORD SizeBufferOut,
    OUT DIOC_GENERATE_MAC_OUT*  DIOCBufferOut
);


// =========================================================================
// =========================================================================

#endif

